@@ -1,14 +1,14 @@
 //! Utilities related to displaying/formatting the edits computed as the difference between two
 //! ASTs
 
-use crate::diff::{Hunk, Hunks, Line};
+use crate::diff::{Hunk, Line, RichHunk, RichHunks};
 use anyhow::Result;
 use console::{Color, Style, Term};
 use log::{debug, info};
 use logging_timer::time;
 use serde::{Deserialize, Serialize};
 use std::{
-    cmp::{max, Ordering},
+    cmp::max,
     io::{BufWriter, Write},
 };
 use strum_macros::EnumString;
@@ -62,7 +62,7 @@ impl Default for ColorDef {
 /// This was abstracted out because the exact same settings apply for both additions and deletions
 #[derive(Debug, Serialize, Deserialize, Eq, PartialEq, Clone)]
 #[serde(rename_all = "kebab-case")]
-pub struct TextFormatting {
+pub struct Config {
     /// The highlight/background color to use with emphasized text
     #[serde(with = "opt_color_def", default = "default_option")]
     pub highlight: Option<Color>,
@@ -96,16 +96,16 @@ struct RegularStyle(Style);
 #[derive(Clone, Debug, PartialEq, Eq)]
 struct EmphasizedStyle(Style);
 
-impl From<&TextFormatting> for RegularStyle {
-    fn from(fmt: &TextFormatting) -> Self {
+impl From<&Config> for RegularStyle {
+    fn from(fmt: &Config) -> Self {
         let mut style = Style::default();
         style = style.fg(fmt.regular_foreground);
         RegularStyle(style)
     }
 }
 
-impl From<&TextFormatting> for EmphasizedStyle {
-    fn from(fmt: &TextFormatting) -> Self {
+impl From<&Config> for EmphasizedStyle {
+    fn from(fmt: &Config) -> Self {
         let mut style = Style::default();
         style = style.fg(fmt.emphasized_foreground);
 
@@ -131,15 +131,15 @@ impl From<&TextFormatting> for EmphasizedStyle {
 #[serde(default, rename_all = "kebab-case")]
 pub struct DiffWriter {
     /// The formatting options to use with text addition
-    pub addition: TextFormatting,
+    pub addition: Config,
     /// The formatting options to use with text addition
-    pub deletion: TextFormatting,
+    pub deletion: Config,
 }
 
 impl Default for DiffWriter {
     fn default() -> Self {
         DiffWriter {
-            addition: TextFormatting {
+            addition: Config {
                 regular_foreground: Color::Green,
                 emphasized_foreground: Color::Green,
                 highlight: None,
@@ -147,7 +147,7 @@ impl Default for DiffWriter {
                 underline: false,
                 prefix: "+ ".into(),
             },
-            deletion: TextFormatting {
+            deletion: Config {
                 regular_foreground: Color::Red,
                 emphasized_foreground: Color::Red,
                 highlight: None,
@@ -162,6 +162,8 @@ impl Default for DiffWriter {
 /// User supplied parameters that are required to display a diff
 #[derive(Debug, Clone, PartialEq)]
 pub struct DisplayParameters<'a> {
+    /// The hunks constituting the diff.
+    pub hunks: RichHunks<'a>,
     /// The parameters that correspond to the old document
     pub old: DocumentDiffData<'a>,
     /// The parameters that correspond to the new document
@@ -173,8 +175,6 @@ pub struct DisplayParameters<'a> {
 pub struct DocumentDiffData<'a> {
     /// The filename of the document
     pub filename: &'a str,
-    /// The edit hunks for the document
-    pub hunks: &'a Hunks<'a>,
     /// The full text of the document
     pub text: &'a str,
 }
@@ -189,8 +189,8 @@ struct FormattingDirectives<'a> {
     pub prefix: &'a dyn AsRef<str>,
 }
 
-impl<'a> From<&'a TextFormatting> for FormattingDirectives<'a> {
-    fn from(fmt_opts: &'a TextFormatting) -> Self {
+impl<'a> From<&'a Config> for FormattingDirectives<'a> {
+    fn from(fmt_opts: &'a Config) -> Self {
         Self {
             regular: fmt_opts.into(),
             emphasis: fmt_opts.into(),
@@ -206,7 +206,7 @@ impl DiffWriter {
     /// between lines.
     #[time("info", "formatting::{}")]
     pub fn print(&self, term: &mut BufWriter<Term>, params: &DisplayParameters) -> Result<()> {
-        let DisplayParameters { old, new } = &params;
+        let DisplayParameters { hunks, old, new } = &params;
         let old_fmt = FormattingDirectives::from(&self.deletion);
         let new_fmt = FormattingDirectives::from(&self.addition);
 
@@ -218,52 +218,15 @@ impl DiffWriter {
 
         self.print_title(term, old.filename, new.filename, &old_fmt, &new_fmt)?;
 
-        // Iterate through the edits on both documents. We know that both of the vectors are
-        // sorted, and we can use that property to iterate through the entries in O(n).
-        let mut it_old = 0;
-        let mut it_new = 0;
-
-        while it_old < old.hunks.0.len() && it_new < new.hunks.0.len() {
-            let old_hunk = &old.hunks.0[it_old];
-            let new_hunk = &new.hunks.0[it_new];
-
-            // We can unwrap here because the loop invariant enforces that there is at least one
-            // element in the deque, otherwise the loop wouldn't run at all.
-            let old_line_num = old_hunk.first_line().unwrap();
-            let new_line_num = new_hunk.first_line().unwrap();
-
-            match old_line_num.cmp(&new_line_num) {
-                Ordering::Equal => {
-                    self.print_hunk(term, &old_lines, old_hunk, &old_fmt)?;
-                    self.print_hunk(term, &new_lines, new_hunk, &new_fmt)?;
-                    it_old += 1;
-                    it_new += 1;
+        for hunk_wrapper in &hunks.0 {
+            match hunk_wrapper {
+                RichHunk::Old(hunk) => {
+                    self.print_hunk(term, &old_lines, hunk, &old_fmt)?;
                 }
-                Ordering::Less => {
-                    self.print_hunk(term, &old_lines, old_hunk, &old_fmt)?;
-                    it_old += 1;
+                RichHunk::New(hunk) => {
+                    self.print_hunk(term, &new_lines, hunk, &new_fmt)?;
                 }
-                Ordering::Greater => {
-                    self.print_hunk(term, &new_lines, new_hunk, &new_fmt)?;
-                    it_new += 1;
-                }
-            };
-        }
-
-        debug!("Printing remaining old hunks");
-
-        while it_old < old.hunks.0.len() {
-            let hunk = &old.hunks.0[it_old];
-            self.print_hunk(term, &old_lines, hunk, &old_fmt)?;
-            it_old += 1;
-        }
-
-        debug!("Printing remaining new hunks");
-
-        while it_new < new.hunks.0.len() {
-            let hunk = &new.hunks.0[it_new];
-            self.print_hunk(term, &new_lines, hunk, &new_fmt)?;
-            it_new += 1;
+            }
         }
         Ok(())
     }
@@ -280,8 +243,6 @@ impl DiffWriter {
         old_fmt: &FormattingDirectives,
         new_fmt: &FormattingDirectives,
     ) -> std::io::Result<()> {
-        let divider = " -> ";
-
         // The different ways we can stack the title
         #[derive(Debug, Eq, PartialEq, PartialOrd, Ord, strum_macros::Display)]
         #[strum(serialize_all = "snake_case")]
@@ -289,6 +250,7 @@ impl DiffWriter {
             Vertical,
             Horizontal,
         }
+        let divider = " -> ";
 
         // We construct the fully horizontal title string. If wider than the terminal, then we
         // format another title string that's vertically stacked
@@ -419,16 +381,16 @@ impl DiffWriter {
         // First, we print the prefix to stdout
         write!(term, "{}", regular.apply_to(fmt.prefix.as_ref()))?;
 
-        // The number of characters that have been printed out to stdout already. These aren't
-        // *actually* chars because UTF-8, but you get the gist.
+        // The number of characters that have been printed out to stdout already. All indices are
+        // in raw byte offsets, as splitting on graphemes, etc was taken care of when processing
+        // the AST nodes.
         let mut printed_chars = 0;
 
         // We keep printing ranges until we've covered the entire line
         for entry in &line.entries {
             // The range of text to emphasize
             // TODO(afnan) deal with ranges spanning multiple rows
-            let emphasis_range =
-                entry.reference.start_position().column..entry.reference.end_position().column;
+            let emphasis_range = entry.start_position().column..entry.end_position().column;
 
             // First we need to see if there's any regular text to cover. If the range has a len of
             // zero this is a no-op
@@ -500,6 +462,7 @@ mod opt_color_def {
     use super::{Color, ColorDef};
     use serde::{Deserialize, Deserializer, Serialize, Serializer};
 
+    #[allow(clippy::trivially_copy_pass_by_ref)]
     pub fn serialize<S>(value: &Option<Color>, serializer: S) -> Result<S::Ok, S::Error>
     where
         S: Serializer,
