@@ -1,17 +1,20 @@
-mod ast;
 mod cli;
 mod config;
 mod diff;
 mod formatting;
+mod input_processing;
 mod neg_idx_vec;
 mod parse;
 
+use crate::parse::supported_languages;
 use anyhow::Result;
-use ast::{AstVector, AstVectorData};
-use cli::{list_supported_languages, set_term_colors, Args};
-use config::{Config, ConfigReadError};
+use clap::IntoApp;
+use clap::Parser;
+use cli::{Args, ColorOutputPolicy};
+use config::{Config, ReadError};
 use console::Term;
 use formatting::{DisplayParameters, DocumentDiffData};
+use input_processing::VectorData;
 use log::{debug, error, info, warn, LevelFilter};
 use parse::{generate_language, language_from_ext, GrammarConfig};
 use serde_json as json;
@@ -50,14 +53,14 @@ fn derive_config(args: &Args) -> Result<Config> {
             // If it is a recoverable error, ex: not being able to find the default file path or
             // not finding a file at all isn't a hard error, it makes sense for us to use the
             // default config.
-            ConfigReadError::ReadFileFailure(_) | ConfigReadError::NoDefault => {
+            ReadError::ReadFileFailure(_) | ReadError::NoDefault => {
                 warn!("{} - falling back to default config", e);
                 Ok(Config::default())
             }
             // If we *do* find a config file and it doesn't parse correctly, we should return an
             // error and let the user know that their config is incorrect. This isn't a browser,
             // we can't just silently march forward and hope for the best.
-            ConfigReadError::DeserializationFailure(e) => {
+            ReadError::DeserializationFailure(e) => {
                 error!("Failed to deserialize config file: {}", e);
                 Err(anyhow::anyhow!(e))
             }
@@ -76,7 +79,7 @@ fn generate_ast_vector_data(
     path: PathBuf,
     file_type: Option<&str>,
     grammar_config: &GrammarConfig,
-) -> Result<AstVectorData> {
+) -> Result<VectorData> {
     let text = fs::read_to_string(&path)?;
     let file_name = path.to_string_lossy();
     debug!("Reading {} to string", file_name);
@@ -90,20 +93,7 @@ fn generate_ast_vector_data(
         info!("Will deduce filetype from file extension");
     };
     let tree = parse::parse_file(&path, file_type, grammar_config)?;
-    Ok(AstVectorData { text, tree, path })
-}
-
-/// Generate an AST vector from the underlying data
-///
-/// This is split off into a function so we can handle things like logging and keep the code DRY
-fn generate_ast_vector(data: &AstVectorData) -> AstVector<'_> {
-    let ast_vec = AstVector::from_ts_tree(&data.tree, &data.text);
-    info!(
-        "Constructed a diff vector with {} nodes for {}",
-        ast_vec.len(),
-        data.path.to_string_lossy(),
-    );
-    ast_vec
+    Ok(VectorData { text, tree, path })
 }
 
 /// Check if the input files are supported by this program.
@@ -114,7 +104,7 @@ fn generate_ast_vector(data: &AstVectorData) -> AstVector<'_> {
 ///
 /// This is used to determine whether the program should fall back to another diff utility.
 fn are_input_files_supported(args: &Args, config: &Config) -> bool {
-    let paths = vec![&args.old, &args.new];
+    let paths = [&args.old, &args.new];
 
     // If there's a user override at the command line, that takes priority over everything else.
     if let Some(file_type) = &args.file_type {
@@ -161,27 +151,30 @@ fn run_diff(args: &Args, config: &Config) -> Result<()> {
 
     let ast_data_a = generate_ast_vector_data(path_a.clone(), file_type, &config.grammar)?;
     let ast_data_b = generate_ast_vector_data(path_b.clone(), file_type, &config.grammar)?;
+    let diff_vec_a = config
+        .input_processing
+        .process(&ast_data_a.tree, &ast_data_a.text);
+    let diff_vec_b = config
+        .input_processing
+        .process(&ast_data_b.tree, &ast_data_b.text);
 
-    let diff_vec_a = generate_ast_vector(&ast_data_a);
-    let diff_vec_b = generate_ast_vector(&ast_data_b);
-
-    let (old_hunks, new_hunks) = ast::compute_edit_script(&diff_vec_a, &diff_vec_b);
+    let hunks = diff::compute_edit_script(&diff_vec_a, &diff_vec_b)?;
     let params = DisplayParameters {
+        hunks,
         old: DocumentDiffData {
             filename: &ast_data_a.path.to_string_lossy(),
-            hunks: &old_hunks,
             text: &ast_data_a.text,
         },
         new: DocumentDiffData {
             filename: &ast_data_b.path.to_string_lossy(),
-            hunks: &new_hunks,
             text: &ast_data_b.text,
         },
     };
-    // Use a buffered terminal instead of a normal unbuffered terminal so we can amortize the cost of printing. It
-    // doesn't really how frequently the terminal prints to stdout because the user just cares about the output at the
-    // end, we don't care about how frequently the terminal does partial updates or anything like that. If the user is
-    // curious about progress, they can enable logging and see when hunks are processed and written to the buffer.
+    // Use a buffered terminal instead of a normal unbuffered terminal so we can amortize the cost
+    // of printing. It doesn't really matter how frequently the terminal prints to stdout because
+    // the user just cares about the output at the end, we don't care about how frequently the
+    // terminal does partial updates or anything like that. If the user is curious about progress,
+    // they can enable logging and see when hunks are processed and written to the buffer.
     let mut buf_writer = BufWriter::new(Term::stdout());
     config.formatting.print(&mut buf_writer, &params)?;
     // Just in case we forgot to flush anything in the `print` function
@@ -217,9 +210,46 @@ fn diff_fallback(cmd: &str, old: &Path, new: &Path) -> io::Result<Child> {
     Command::new(cmd).args([old, new]).spawn()
 }
 
-#[paw::main]
-fn main(args: Args) -> Result<()> {
+/// Print a list of the languages that this instance of diffsitter was compiled with
+pub fn list_supported_languages() {
+    #[cfg(feature = "static-grammar-libs")]
+    {
+        let languages = supported_languages();
+        println!("This program was compiled with support for:");
+
+        for language in languages {
+            println!("- {}", language);
+        }
+    }
+
+    #[cfg(feature = "dynamic-grammar-libs")]
+    {
+        println!("This program will dynamically load grammars from shared libraries");
+    }
+}
+
+/// Set whether the terminal should display colors based on the user's preferences
+///
+/// This method will set the terminal output policy *for the current thread*.
+fn set_term_colors(color_opt: ColorOutputPolicy) {
+    match color_opt {
+        ColorOutputPolicy::Off => (console::set_colors_enabled(false)),
+        ColorOutputPolicy::On => (console::set_colors_enabled(true)),
+        ColorOutputPolicy::Auto => (),
+    };
+}
+
+/// Print shell completion scripts to `stdout`.
+///
+/// This is a basic wrapper for the subcommand.
+fn print_shell_completion(shell: clap_complete::Shell) {
+    let mut app = cli::Args::command();
+    clap_complete::generate(shell, &mut app, "diffsitter", &mut io::stdout());
+}
+
+fn main() -> Result<()> {
     use cli::Command;
+    let args = Args::parse();
 
     // We parse the config as early as possible so users can get quick feedback if anything is off
     // with their config.
@@ -233,6 +263,9 @@ fn main(args: Args) -> Result<()> {
 
             #[cfg(feature = "better-build-info")]
             Command::BuildInfo => print_build_info(),
+            Command::GenCompletion { shell } => {
+                print_shell_completion(shell.into());
+            }
         }
     } else {
         let log_level = if args.debug {
@@ -244,7 +277,6 @@ fn main(args: Args) -> Result<()> {
             .filter_level(log_level)
             .init();
         set_term_colors(args.color_output);
-
         // First check if the input files can be parsed with tree-sitter.
         let files_supported = are_input_files_supported(&args, &config);
 
@@ -283,20 +315,32 @@ mod tests {
             "test data path {} does not exist",
             path_b.to_str().unwrap()
         );
+
         (path_a, path_b)
     }
 
-    #[test_case("short", "rust", "rs")]
-    #[test_case("short", "python", "py")]
-    #[test]
-    fn diff_hunks_snapshot(test_type: &str, name: &str, ext: &str) {
+    #[test_case("short", "rust", "rs", true)]
+    #[test_case("short", "python", "py", true)]
+    #[test_case("medium", "rust", "rs", true)]
+    #[test_case("medium", "rust", "rs", false)]
+    #[test_case("medium", "cpp", "cpp", true)]
+    #[test_case("medium", "cpp", "cpp", false)]
+    fn diff_hunks_snapshot(test_type: &str, name: &str, ext: &str, split_graphemes: bool) {
         let (path_a, path_b) = get_test_paths(test_type, name, ext);
         let config = GrammarConfig::default();
         let ast_data_a = generate_ast_vector_data(path_a, None, &config).unwrap();
         let ast_data_b = generate_ast_vector_data(path_b, None, &config).unwrap();
-        let diff_vec_a = generate_ast_vector(&ast_data_a);
-        let diff_vec_b = generate_ast_vector(&ast_data_b);
-        let diff_hunks = ast::compute_edit_script(&diff_vec_a, &diff_vec_b);
-        assert_debug_snapshot!(diff_hunks);
+
+        let processor = input_processing::TreeSitterProcessor { split_graphemes };
+
+        let diff_vec_a = processor.process(&ast_data_a.tree, &ast_data_a.text);
+        let diff_vec_b = processor.process(&ast_data_b.tree, &ast_data_b.text);
+        let diff_hunks = diff::compute_edit_script(&diff_vec_a, &diff_vec_b);
+
+        // We have to set the snapshot name manually, otherwise there appear to be threading issues
+        // and we end up with more snapshot files than there are tests, which cause
+        // nondeterministic errors.
+        let snapshot_name = format!("{test_type}_{name}_{split_graphemes}");
+        assert_debug_snapshot!(snapshot_name, diff_hunks);
     }
 }
